/* 
 * File:   ProvaServo_main.c
 * Author: matteo
 *
 * Created on 6 marzo 2012, 23.20
 */

/*PROVA SERVOMOTORI A 4 USCITE                                                 *
 *tale programma consente di pilotare fino a 4 servomotori alla posizione      *
 *voluta con risoluzione di 0,05 ms, corrispondenti a circa 20 step con        *
 *servomotori standard                                                         *
 *Ogni motore  abilitabile e regolabile indipendentemente                     *
 *Le connessioni elettriche sono stabilite nel file "DefaultValues.h" come     *
 *anche i tempi min/max e la pausa tra un impulso e l'altro                    *
 ******************************************************************************
 *USO: all'accensione tutti i servo sono disattivati. Agendo sul pulsante "NX" *
 *       possibile selezionare i servomotori, il simbolo '*' sul display      *
 *      indicher il servo selezionato.                                        *
 *      Per abilitare/disabilitare il servo premere il pulsante "EN"           *
 *      Per modificarne la posizione usare i tasti "UP" e "DN"                 *
 *******************************************************************************
 *POSSIBILI SVILUPPI DELL'APPLICAZIONE:                                        *
 * Il codice  scritto in modo da garantire l'espandibilit del sistema,      *
 * pertanto possibile aumentare il numero di servomotori controllabili,        *
 * teoricamente fino a saturare tutte i pin I/O del pic usato                  *
 *                                                                             *
 * Il programma funziona su qualsiasi PIC dotato della risorsa TIMER0 e con    *
 * clock di 20MHz, potrebbe esser necessario ridefinire alcuni valori          *
 * riguardanti l'hardware. Ovviamente va compilato per il pic che si intende   *
 * utilizzare.                                                                 *
 *                                                                             *
 * Si pu provare a ridurre il tempo di interrupt per aumentare la risoluzione *
 * e quindi la precisione del controllo.                                       *
 *                                                                             *
 * Un'altra modifica potr esser l'implementazione di una routine di comando   *
 * tramite seriale, per controllare il dispositivo tramite PC                  *
 *                                                                             *
 *******************************************************************************/

/*ATTENZIONE!!                                                                 *
 *si consiglia di collegare il servomotore tramite resistenza ai pin del pic   *
 *valore intorno ai 470ohm                                                     *
 *posizionare un diodo in antiparallelo all'alimentaizione del motore          *
 *molto probabilmente servir di usare un regolatore di tensione separato per  *
 *alimentare il motore                                                         *
 *******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <htc.h>
#include <freedomII_hardware.h>
#include "DefaultValues.h"
#include <delay/delay.c>
#include <lcd.c>
__CONFIG(WDTE_OFF & LVP_OFF & FOSC_HS & DEBUG_OFF);


void SFRsetup(void); //impostazioni SFR

void SERVO(unsigned char s1, unsigned char s2, unsigned char s3, unsigned char s4);
unsigned char servo_counter;

void BTNhandler();
unsigned int btn_counter;
unsigned char btn_busy;
static bit newvalue;
#define UP_busy 1
#define DN_busy 2
#define EN_busy 3
#define NX_busy 4

void LCDstart();
void LCDwriter();
unsigned int lcd_counter;

static bit main_int;

unsigned char servo1=servo_c;
unsigned char servo2=servo_c;
unsigned char servo3=servo_c;
unsigned char servo4=servo_c;
static bit en1;
static bit en2;
static bit en3;
static bit en4;

unsigned char servo_sel=1;


void interrupt isr() //Routine di interrupt
{
    testpin = !testpin; /*inverte lo stato logico su testpin per poter misurare*
                         *il tempo tra un interrupt e l'altro, utile per tarare*
                         *il preload del TIMER 0                               */
    
    TMR0=TMR0_preload; //ricarica il TIMER 0 con il valore di preload definito
    main_int = 1;      //setta il flag di avvenuto interrupt
    TMR0IF=0;          //azzera il flag interrupt TIMER 0

}

void SFRsetup() //Routine di impostazione all'avvio Special Function Register
{
        //setup PORTA
    TRISA=0x00;
    PORTA=0x00;
    //setup PORTB
    TRISB=0xF0;
    PORTB=0x00;
    //setup PORTC
    TRISC=0x00;
    PORTC=0x00;
    //setup PORTD
    TRISD=0x00;
    PORTD=0x00;
    //setup PORTE
    TRISE=0x00;
    PORTE=0x00;

    OPTION_REG=0b01001000;
    //RBPU    0     pullup PORTB attivati
    //INTEDG  1     *interrupt RB0 su fronte SALITA
    //T0CS    0     clock TMR0 su ciclo istruzioni interno
    //T0SE    0     (solo extCK) incremento TMR0 su fronte salita RA4/T0CKI
    //PSA     1     prescaler assegnato al WDT
    //PS2:PS0 000   prescaler 1:1 WDT

    INTCON=0b10100000;
    //GIE       1   Interrupt abilitati
    //PEIE      0   Interrupt periferiche disabilitati
    //TMR0IE    1   Interrupt TMR0 abilitato
    //INTE      0   Interrupt RB0/INT disabilitato
    //RBIE      0   Interrupt PORTB disabilitato
    //TMR0IF    x   Flag Interrupt TMR0
    //INTF      x   Flag Interrupt RB0/INT
    //RBIF      x   Flag Interrupt PORTB
}

void SERVO(unsigned char s1, unsigned char s2, unsigned char s3, unsigned char s4)
/*Routine di comando servomotori.                                              *
 *La routine  eseguita ad ogni interrupt di TMR0. Usa un contatore interno    *
 *incrementato ad ogni esecuzione (servo_counter), e tramite confronto con il  *
 *valore di tempo desiderato stabilisce se l'uscita di un servo deve esser alta*
 *oppure bassa.                                                                *
 *Utilizza i seguenti define e variabili esterne:                              *
 * out1...4 : pin su cui vengono generati gli impulsi                          *
 * en1...4 : bit di enable dei rispettivi pin                                  *
 * servo_pause : tempo tra l'inizio di un impulso ed il successivo             *
 *******************************************************************************
 * Le variabili di ingresso sono:                                              *
 * s1...s4 : tempo desiderato per ogni servo                                   *
 *******************************************************************************/
{
    if(en1 && servo_counter < s1){out1 = 1;}
    else{out1 = 0;}

    if(en2 && servo_counter < s2){out2 = 1;}
    else{out2 = 0;}

    if(en3 && servo_counter < s3){out3 = 1;}
    else{out3 = 0;}

    if(en4 && servo_counter < s4){out4 = 1;}
    else{out4=0;}

    servo_counter++;
    if(servo_counter > servo_pause){servo_counter=0;}
}

void BTNhandler()
/*Routine di gestione dei pulsanti                                             *
 *Tale routine gestisce l'input utente tramite pulsanti, occupandosi di        *
 *effettuare le funzioni di antirimbalzo e l'esecuzione dei comandi associati  *
 *ai pulsanti                                                                  *
 *Se vi  un pulsante premuto viene incrementato un contatore (btn_counter)    *
 *e viene impostato un valore sulla variabile btn_busy per indicare qual il   *
 *pulsante premuto. Il contatore  incrementato ad ogni ciclo programma (TMR0) *
 *Se il pulsante rimane premuto un tempo sufficiente definito da BOUNCE,       *
 *calcolabile come: Tempo interrupt * BOUNCE (in questo caso 0,05ms*BOUNCE),   *
 *allora al rilascio del pulsante viene eseguito l'evento associato allo stesso*
 *I pulsanti 1 e 2 (qui identificati con UP e DN) prevedono una modalit di    *
 *ripetizione rapida del comando, il tempo di ripetizione  definito da        *
 *BTNrepeat, calcolabile sempre come: Tempo interrupt * BTNrepeat              *
 *******************************************************************************
 *Funzioni svolte dai pulsanti:                                                *
 *UP: incrementa la larghezza dell'impulso del servo selezionato               *
 *DN: riduce la larghezza dell'impulso del servo selezionato                   *
 *EN: Attiva/Disattiva il servo selezionato                                    *
 *NX: Seleziona il servo successivo                                            *
 *******************************************************************************/
{
    if(!UP) //PRESSIONE PULSANTE INCREMENTO
    {
        btn_counter++; //Incremento il contatore di antirimbalzo
        btn_busy = UP_busy; //Imposto btn_busy = pulsante UP
        if(btn_counter++ > BTNrepeat){goto btnUP;} /*Verifico che sia trascorso*
                                                    *il tempo di ripetizione   *
                                                    *del comando, in caso      *
                                                    *positivo passo ad eseguire*
                                                    *il comando associato      **/
    }
    else if(btn_busy==UP_busy && btn_counter>=BOUNCE) /*Al rilascio del        *
                                                       *pulsante verifico che  *
                                                       *sia trascorso il tempo *
                                                       *di debounce, in caso   *
                                                       *positivo eseguo il     *
                                                       *comando associato.     *
                                                       *Il controllo di        *
                                                       *btn_busy permette di   *
                                                       *determinare il pulsante*
                                                       *premuto                */
    {
        btnUP: /*comandi eseguiti dal pulsante UP: si occupa di incrementare la*
                *larghezza dell'impulso associato ad un servomotore,           *
                *incrementandone la variabile associata (es. servo1).          *
                *In base al servo selezionato incrementa la relativa variabile *
                *fino al valore massimo definito da "servo_max"                */
        switch (servo_sel)
        {
            case 1:
                if(servo1<servo_max){servo1++;}
                break;
            case 2:
                if(servo2<servo_max){servo2++;}
                break;
            case 3:
                if(servo3<servo_max){servo3++;}
                break;
            case 4:
                if(servo4<servo_max){servo4++;}
        }

        btn_counter=0; //azzero il contatore di debounce
        btn_busy=0; //reimposto il pulsante selezionato a 0 (ovvero nessuno)
        newvalue=1; /*imposto il flag di nuovo valore affinch venga aggiornato*
                     *il display dall'apposita routine                         */
    }

    if(!DN) //PRESSIONE PULSANTE DECREMENTO
    {
        btn_counter++;
        btn_busy = DN_busy;
        if(btn_counter++ > BTNrepeat){goto btnDN;}
    }
    else if(btn_busy==DN_busy && btn_counter>=BOUNCE)
    {
        btnDN: /*comandi eseguiti dal pulsante DN: si occupa di ridurre la     *
                *larghezza dell'impulso associato ad un servomotore,           *
                *decrementandone la variabile associata (es. servo1).          *
                *In base al servo selezionato decrementa la relativa variabile *
                *fino al valore minimo definito da "servo_min"                 */
        switch (servo_sel)
        {
            case 1:
                if(servo1>=servo_min){servo1--;}
                break;
            case 2:
                if(servo2>=servo_min){servo2--;}
                break;
            case 3:
                if(servo3>=servo_min){servo3--;}
                break;
            case 4:
                if(servo4>=servo_min){servo4--;}
                break;
        }

        btn_counter=0;
        btn_busy=0;
        newvalue=1;
    }

    if(!EN) //PRESSIONE PULSANTE ENABLE
    {
        btn_counter++;
        btn_busy = EN_busy;
    }
    else if(btn_busy==EN_busy && btn_counter>=BOUNCE)
    {
        switch (servo_sel) /*comandi eseguiti dal pulsante EN: attiva/disattiva*
                            *il servo selezionato, settandone il relativo flag */
        {
            case 1: en1 = !en1; break;
            case 2: en2 = !en2; break;
            case 3: en3 = !en3; break;
            case 4: en4 = !en4; break;
        }
        btn_counter=0;
        btn_busy=0;
        newvalue=1;
    }

    if(!NX) //PRESSIONE PULSANTE SELEZIONE SUCCESSIVO
    {
        btn_counter++;
        btn_busy = NX_busy;
    }
    else if(btn_busy==NX_busy && btn_counter>=BOUNCE)
    {
                        /*comandi eseguiti dal pulsante NX: seleziona il servo *
                         *successivo incrementando la variabile servo_sel      *
                         *usata per definire il servo selezionato nelle routine*
                         *degli altri pulsanti                                 */
        servo_sel++;
        if(servo_sel>4){servo_sel=1;}

        btn_counter=0;
        btn_busy=0;
        newvalue=1;
    }
    
}

void LCDstart() //scrive l'LCD all'accensione con i valori predefiniti
{
    LCD_INIT();
    LCD_CUSTOMCHAR(0, 0,0,0,0,0,0,0,0);
    LCD_CLEAR();
    LCD_PUTS("s1:off*   s2:off");
    LCD_GOTO(2,1);
    LCD_PUTS("s3:off    s4:off");
}

void LCDwriter() //Aggiorna la visualizzazione quando vengono modificati dei valori
{
    switch (servo_sel)
    {
        case 1:
            LCD_GOTO(2,10);LCD_PUTCH(' ');
            LCD_GOTO(1,4);
            if(en1)
            {
                if(servo1<100){LCD_PUTCH(' ');}
                if(servo1<10){LCD_PUTCH(' ');}
                LCD_PUTUN(servo1);
            }
            else
            {
                LCD_PUTS("off");
            }
            LCD_PUTCH('*');
            break;

        case 2:
            LCD_GOTO(1,7);
            LCD_PUTS("   *");
            LCD_GOTO(1,14);
            if(en2)
            {
                if(servo2<100){LCD_PUTCH(' ');}
                if(servo2<10){LCD_PUTCH(' ');}
                LCD_PUTUN(servo2);
            }
            else
            {
                LCD_PUTS("off");
            }
            break;

        case 3:
            LCD_GOTO(1,10);LCD_PUTCH(' ');
            LCD_GOTO(2,4);
            if(en3)
            {
                if(servo3<100){LCD_PUTCH(' ');}
                if(servo3<10){LCD_PUTCH(' ');}
                LCD_PUTUN(servo3);
            }
            else
            {
                LCD_PUTS("off");
            }
            LCD_PUTCH('*');
            break;

        case 4:
            LCD_GOTO(2,7);
            LCD_PUTS("   *");
            LCD_GOTO(2,14);
            if(en4)
            {
                if(servo4<100){LCD_PUTCH(' ');}
                if(servo4<10){LCD_PUTCH(' ');}
                LCD_PUTUN(servo4);
            }
            else
            {
                LCD_PUTS("off");
            }
            break;

    }
}



/*INIZIO PROGRAMMA PRINCIPALE*/
void main()
{
    SFRsetup(); //Inizializza gli Special Function Registers

    LCDstart(); //Inizializza l'LCD

    while(1)    //Ciclo infinito
    {
        if(main_int) //Verifico che sia avvenuto l'interrupt principale (TMR0)
        {
            BTNhandler(); //Richiamo la routine di gestione pulsanti
            SERVO(servo1,servo2,servo3,servo4); //Richiamo la routine di gestione motori

            if(newvalue) /*Verifico se sono stati modificati dei valori        *
                          * in caso positivo provvedo a richiamare la routine  *
                          * di gestione Display, viceversa salta per non       *
                          * impiegare tempo inutilmente. Dopo aver richiamato  *
                          * la routine il flag (newvalue) viene azzerato       */
            {
                LCDwriter();
                newvalue=0;
            }


            main_int=0; /*azzero il flag dell'interrupt per poter intercettare *
                         *un nuovo interrupt                                   */
        }
    }


}